# -*- coding: utf-8 -*-

"""
Module for collecting configuration data from ``~/sen2chain_data/config/config.cfg``
"""

import os
import logging
from pathlib import Path
from configparser import ConfigParser


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

ROOT = Path(os.path.realpath(__file__)).parent.parent

SHARED_DATA = dict(
    tiles_index=ROOT / "sen2chain" / "data" / "tiles_index.gpkg",
    tiles_index_dict=ROOT / "sen2chain" / "data" / "tiles_index_dict.p",
    peps_download=ROOT / "sen2chain" / "peps_download3.py")


class Config:
    """Class for loading, checking, and retrieving configuration parameters.
    The class reads during its first initialization the configurable settings
    from the configuration file.

    Usage::
        >>> Config().get("l1c_path")
    """
    # TODO: Implement the Config class as a singleton.

    _USER_DIR = Path.home() / "sen2chain_data"
    _CONFIG_DIR = _USER_DIR / "config"
    _DEFAULT_DATA_DIR = _USER_DIR / "data"
    _CONFIG_FILE = _CONFIG_DIR / "config.cfg"
    _TILES_TO_WATCH = _CONFIG_DIR / "tiles_to_watch.csv"

    def __init__(self) -> None:
        self._config_params = {"temp_path", "l1c_path", "l2a_path",
                               "indices_path", "time_series_path",
                               "sen2cor_bashrc_path", "scihub_id",
                               "scihub_pwd", "peps_config_path"}
        self.cfg = None
        self.config_dict = dict()

        if not self._CONFIG_FILE.exists():
            self._create_config()

        self._load_config()
        self._check_data_paths()

    def _create_config(self) -> None:
        """Create a new config file in ``~/senchain_data/config/config.cfg``."""
        self._USER_DIR.mkdir(exist_ok=True)
        self._CONFIG_DIR.mkdir(exist_ok=True)
        self._DEFAULT_DATA_DIR.mkdir(exist_ok=True)

        with open(str(self._CONFIG_FILE), "w") as cfg_file:
            self.cfg = ConfigParser()
            self.cfg.read(str(self._CONFIG_FILE))
            self.cfg.add_section("DATA PATHS")
            self.cfg.add_section("SEN2COR PATH")
            self.cfg.add_section("HUBS LOGINS")
            self.cfg.set("DATA PATHS", "temp_path", "")
            self.cfg.set("DATA PATHS", "l1c_path", "")
            self.cfg.set("DATA PATHS", "l2a_path", "")
            self.cfg.set("DATA PATHS", "indices_path", "")
            self.cfg.set("DATA PATHS", "time_series_path", "")
            self.cfg.set("SEN2COR PATH", "sen2cor_bashrc_path", "")
            self.cfg.set("HUBS LOGINS", "peps_config_path", "")
            self.cfg.set("HUBS LOGINS", "scihub_id", "")
            self.cfg.set("HUBS LOGINS", "scihub_pwd", "")

            self.cfg.write(cfg_file)

    def _load_config(self) -> None:
        """Loads configuration parameters in a dictionnary."""
        self.cfg = ConfigParser()
        self.cfg.read(str(self._CONFIG_FILE))

        # check if a parameter is missing in the config file
        for section in self.cfg.sections():
            for key in self.cfg[section].keys():
                if key in self._config_params:
                    self.config_dict[key] = self.cfg[section][key]
                else:
                    raise ValueError("Invalid parameter:", key)

        if self._config_params != set(self.config_dict.keys()):
            missing_params = self._config_params - set(self.config_dict.keys())
            raise ValueError(
                "Parameter not found in config file: {}".format(missing_params)
                )

    def _check_data_paths(self) -> None:
        """
        Checks if data paths are provided and valids. If not, create default
        folders in sen2chain_data/DATA and update the configuration file.
        """
        def update_config(section, key, val):
            """
            Update a setting in config.ini
            """
            self.cfg.set(section, key, val)
            with open(str(self._CONFIG_FILE), "w") as cfg_file:
                self.cfg.write(cfg_file)

        data_paths = ["temp_path", "l1c_path", "l2a_path",
                      "indices_path", "time_series_path"]

        for path in data_paths:
            value = self.config_dict[path]

            if value.rstrip() == "" or not Path(value).exists():

                default_value = self._DEFAULT_DATA_DIR / path.replace("_path", "").upper()
                default_value.mkdir(parents=True, exist_ok=True)
                update_config("DATA PATHS", path, str(default_value))

                logger.info("{}: using default at {}".format(path, str(default_value)))

        sen2cor_bashrc_path_value = self.config_dict["sen2cor_bashrc_path"]

        if sen2cor_bashrc_path_value.rstrip() == "" or not Path(sen2cor_bashrc_path_value).exists():
            logging.error("Make sure the path to the sen2cor Bashrc file is valid.")
            raise ValueError("Invalid sen2cor Bashrc")

    def get(self, param: str) -> str:
        """Returns a parameter value.

        :param param: parameter.
        """
        if param not in self._config_params:
            raise ValueError("Invalid parameter", param)

        return self.config_dict.get(param)

    @property
    def tiles_to_watch(self):
        return self._TILES_TO_WATCH
