# -*- coding: utf-8 -*-

"""
This module contains functions to compute radiometric indices.
"""

import logging
import pathlib
from pathlib import Path
import otbApplication
import rasterio
from rasterio.warp import reproject, Resampling
import numpy as np
from typing import Union
import gdal

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def create_raw_ndvi(
        nir_path: Union[str, pathlib.PosixPath],
        vir_path: Union[str, pathlib.PosixPath],
        out_path: Union[str, pathlib.PosixPath]="./raw_ndvi.tif"
) -> pathlib.PosixPath:
    """
    Creates a NDVI raster from NIR and VIR rasters.

    :param nir_path: path to the NIR raster.
    :param vir_path: path to the VIR raster.
    :param out_path: path to the output raster.
    """
    logger.info("creating raw NDVI (tiff - int16)")
    
    with rasterio.open(str(nir_path)) as nir_src, \
            rasterio.open(str(vir_path)) as vir_src:
        nir_profile = nir_src.profile
        np.seterr(divide='ignore', invalid='ignore')    # ignore warnings when dividing by zero
        nir = nir_src.read(1).astype(np.float32)
        vir = vir_src.read(1).astype(np.float32)
        ndvi = ((nir - vir) / (nir + vir)*10000).astype(np.int16)
        ndvi_masked = np.where(nir != 0, ndvi, 32767)

    nir_profile.update(driver="Gtiff",
                       compress="DEFLATE",
                       tiled=False,
                       dtype=np.int16,
                       nodata=32767,
                       transform=nir_src.transform)
    nir_profile.pop('tiled', None)
    with rasterio.Env(GDAL_CACHEMAX=512) as env:
        with rasterio.open(str(out_path), "w", **nir_profile) as dst:
            dst.write(ndvi_masked, 1)
    return Path(str(out_path)).absolute


def create_raw_ndwi(nir_path: Union[str, pathlib.PosixPath],
        swir_path: Union[str, pathlib.PosixPath],
        out_path: Union[str, pathlib.PosixPath]="./raw_ndwigao.tif") -> pathlib.PosixPath:
    """
    Creates a NDWI raster from NIR and SWIR rasters.

    :param nir_path: path to the NIR raster.
    :param swir_path: path to the SWIR raster.
    :param out_path: path to the output raster.
    """
    logger.info("creating raw NDWIGAO (tiff - int16)")

    with rasterio.open(str(nir_path)) as nir_src, \
            rasterio.open(str(swir_path)) as swir_src:

        nir_profile = nir_src.profile
        # swir_profile = swir_src.profile

        nir = nir_src.read(1).astype(np.float32)
        swir = swir_src.read(1).astype(np.float32)

        # reproject swir band (20m) to nir band resolution (10m)
        swir_reproj = np.empty(nir.shape, dtype=np.float32)
        reproject(source=swir,
                  destination=swir_reproj,
                  src_transform=swir_src.transform,
                  src_crs=swir_src.crs,
                  dst_transform=nir_src.transform,
                  dst_crs=nir_src.crs,
                  resampling=Resampling.nearest)

        np.seterr(divide='ignore', invalid='ignore')    # ignore warnings when dividing by zero
        ndwi = ((nir - swir_reproj) / (nir + swir_reproj)*10000).astype(np.int16)

        ndwi_masked = np.where(nir != 0, ndwi, 32767)

    nir_profile.update(driver="Gtiff",
                       compress="DEFLATE",
                       tiled=False,
                       dtype=np.int16,
                       nodata=32767,
                       transform=nir_src.transform)
    nir_profile.pop('tiled', None)
    with rasterio.Env(GDAL_CACHEMAX=512) as env:
        with rasterio.open(str(out_path), "w", **nir_profile) as dst:
            dst.write(ndwi_masked, 1)
    return Path(str(out_path)).absolute


def create_raw_ndwimcf(nir_path: Union[str, pathlib.PosixPath],
        green_path: Union[str, pathlib.PosixPath],
        out_path: Union[str, pathlib.PosixPath]="./raw_ndwimcf.tif") -> pathlib.PosixPath:
    """
    Creates a NDWI (McFeeters) raster from GREEN and NIR rasters.

    :param nir_path: path to the NIR raster.
    :param green_path: path to the GREEN raster.
    :param out_path: path to the output raster.
    """
    logger.info("creating raw NDWIMCF (tiff - int16)")

    with rasterio.open(str(nir_path)) as nir_src, \
            rasterio.open(str(green_path)) as green_src:

        nir_profile = nir_src.profile
        # swir_profile = swir_src.profile

        nir = nir_src.read(1).astype(np.float32)
        green = green_src.read(1).astype(np.float32)

        # reproject swir band (20m) to nir band resolution (10m)
        #~ swir_reproj = np.empty(nir.shape, dtype=np.float32)
        #~ reproject(source=swir,
                  #~ destination=swir_reproj,
                  #~ src_transform=swir_src.transform,
                  #~ src_crs=swir_src.crs,
                  #~ dst_transform=nir_src.transform,
                  #~ dst_crs=nir_src.crs,
                  #~ resampling=Resampling.nearest)

        np.seterr(divide='ignore', invalid='ignore')    # ignore warnings when dividing by zero
        ndwimcf = ((green - nir) / (green + nir)*10000).astype(np.int16)

        ndwimcf_masked = np.where(nir != 0, ndwimcf, 32767)

    nir_profile.update(driver="Gtiff",
                       compress="DEFLATE",
                       tiled=False,
                       dtype=np.int16,
                       nodata=32767,
                       transform=nir_src.transform)
    nir_profile.pop('tiled', None)
    with rasterio.Env(GDAL_CACHEMAX=512) as env:
        with rasterio.open(str(out_path), "w", **nir_profile) as dst:
            dst.write(ndwimcf_masked, 1)
    return Path(str(out_path)).absolute

def create_masked_indice(
        indice_path: Union[str, pathlib.PosixPath],
        cloud_mask_path: Union[str, pathlib.PosixPath],
        out_path: Union[str, pathlib.PosixPath]="./masked_indice.tif"
    ) -> pathlib.PosixPath:
    """
    Masks an indice raster with a cloud mask.

    :param indice_path: path to the NDVI raster.
    :param cloud_mask_path: path to the cloud mask raster.
    :param out_path: path to the output raster.
    """
    # TODO: Improve the function in order to mask any raster.

    logger.info("Cloud-masking indice (int16)")

    with rasterio.open(str(indice_path)) as indice_src, \
            rasterio.open(str(cloud_mask_path)) as cld_src:

        profile = indice_src.profile
        raw_indice = indice_src.read(1)
        cld = cld_src.read(1)
        # repoject cloud_mask to ndvi resolution
        cld_reproj = np.empty(raw_indice.shape, dtype=np.uint8)
        reproject(source=cld,
                  destination=cld_reproj,
                  src_transform=cld_src.transform,
                  src_crs=cld_src.crs,
                  dst_transform=indice_src.transform,
                  dst_crs=indice_src.crs,
                  resampling=Resampling.nearest)
#        indice_borders_mask = np.where(raw_indice > 0, raw_indice, 32767)
        indice_cloud_mask = np.where(cld_reproj == 0, raw_indice, 32767)

        profile.update(driver="Gtiff",
                        compress="DEFLATE",
                        tiled=False,
                        dtype=np.int16,
                        nodata=32767,
                        transform=indice_src.transform)
#        profile.pop('tiled', None)

    with rasterio.Env(GDAL_CACHEMAX=512) as env:
        with rasterio.open(str(out_path), "w", **profile) as dst:
            dst.write(indice_cloud_mask, 1)

    return str(Path(str(out_path)).absolute)

def index_tiff_2_jp2(img_path: Union[str, pathlib.PosixPath],
        out_path: Union[str, pathlib.PosixPath]="./indice_2_jp2.jp2") -> pathlib.PosixPath:
    """
    Convert a indice file from TIF to JP2.
    :param out_path: path to the output raster.
    """
    logger.info("converting raw indice to JPEG2000")
    src_ds = gdal.Open(str(img_path))
    driver = gdal.GetDriverByName("JP2OpenJPEG")
    for i in range(src_ds.RasterCount):
            src_ds.GetRasterBand(i+1).SetNoDataValue(float(16383))
    dst_ds = driver.CreateCopy(str(out_path), src_ds, options=['NBITS=15', 'CODEC=JP2', 'QUALITY=20'])
    dst_ds = None
    src_ds = None
    return str(Path(str(out_path)).absolute)
