# -*- coding: utf-8 -*-

"""
Module for computing time series.
"""

import logging
import pathlib
from pathlib import Path
from collections import OrderedDict
import fiona
import rasterio
from fiona import transform
from shapely.geometry import shape
from rasterstats import zonal_stats
import pandas as pd
# type annotations
from typing import Sequence, List, Dict, Union

from .config import Config
from .tiles import Tile, ProductsList
from .indices import IndicesCollection
from .geo_utils import get_tiles_from_file


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class TimeSeries:
    """Class for time series extraction.

    :param vectors_file: path to the vector file.
    :param indices: list of valid indices names.
    :param date_min: starting date of the time series (YYYY-MM-DD format).
    :param date_max: ending date of the time series (YYYY-MM-DD format).
    :param cover_min: minimum cloud cover value. Default: 0.
    :param cover_max: minimum cloud cover value. Default: 100.

    Usage:
        >>> ts = TimeSeries("polygons.geojson", indices=["NDVI"], date_min="2017-01-01", date_max="2018-01-01")
    """
    def __init__(
            self,
            vectors_file: Union[str, pathlib.PosixPath],
            indices: Sequence[str] = None,
            date_min: str = None, date_max: str = None,
            cover_min: int = 0, cover_max: int = 100
    ) -> None:

        self._vectors_file = Path(vectors_file)
        self._date_min = date_min
        self._date_max = date_max
        self._cover_min = cover_min
        self._cover_max = cover_max

        if indices is None:
            self._indices_list = IndicesCollection.list
        else:
            for indice in indices:
                if indice not in IndicesCollection:
                    raise ValueError("Invald indice name: {}".format(indice))
            self._indices_list = indices

        self._tiles_geoms = self._get_tiles_geom_dict()
        self._df_dicts = dict()
        self._get_stats()

    @property
    def data(self) -> Dict[str, pd.DataFrame]:
        """Returns a dictionnary with indices as keys and pandas.DataFrame as values."""
        return self._df_dicts

    @staticmethod
    def _is_indice_in_tile(indice: str, tile: Tile) -> bool:
        """Does the tile has a folder for this indice.

        :param indice: indice name.
        :param tile: Tile object.
        """
        if indice.lower() in tile.paths["indices"]:
            return True
        return False

    def _get_tiles_geom_dict(self) -> Dict[str, List[str]]:
        """{Feature id: list of intersected tiles} to {tile: list of intersected features ids}."""
        intersected_tiles_dict = get_tiles_from_file(self._vectors_file, land_only=False)
        tiles_geom = dict()
        for feat_id, tiles_list in intersected_tiles_dict.items():
            for tile in tiles_list:
                if tile not in tiles_geom:
                    tiles_geom[tile] = []
                tiles_geom[tile].append(feat_id)
        return tiles_geom

    @staticmethod
    def _filter_products(
            tile: Tile,
            indice: str,
            date_min: str, date_max: str,
            cover_min: int, cover_max: int
    ) -> ProductsList:
        """Filters tile's indices products.

        :param tile: tile.
        :param indice: indice.
        :param date_min: oldest date.
        :param date_max: most recent date.
        :param cover_min: minimum cloud coverage.
        :param cover_max: maximum cloud coverage.
        """
        products = vars(tile)[indice.lower()].masks
        filt = products.filter_dates(date_min, date_max).filter_clouds(cover_min, cover_max)
        return filt

    @staticmethod
    def _get_raster_stats_in_geom(
            feature: Dict,
            raster_path: Union[str, pathlib.PosixPath]
    ) -> Dict:
        """Extracts statistics from a raster in a geometry.

        :param feature: GeoJSON like object.
        :param raster_path: path to the raster.
        """
        geom = shape(transform.transform_geom("EPSG:4326",
                                              rasterio.open(str(raster_path)).crs["init"],
                                              feature["geometry"]))

        stats = zonal_stats(geom, str(raster_path), stats=["count",
                                                           "nodata",
                                                           "min",
                                                           "max",
                                                           "median",
                                                           "mean",
                                                           "std"])[0]
        return stats

    def _get_stats(self) -> None:
        """Compute stats in polygons."""
        with fiona.open(str(self._vectors_file), "r") as vectors:
            features = {feat["id"]: feat for feat in vectors}

            for indice in self._indices_list:
                logger.info("computing {}".format(indice))
                rows_list = []

                for tile, fid_list in self._tiles_geoms.items():
                    tile_obj = Tile(tile)

                    logger.info("Tile: {}".format(tile))

                    if TimeSeries._is_indice_in_tile(indice, tile_obj):
                        tile_indice_path = tile_obj.paths["indices"][indice.lower()]
                        products = TimeSeries._filter_products(tile_obj, indice,
                                                               self._date_min, self._date_max,
                                                               self._cover_min, self._cover_max)

                        for prod in products:
                            prod_path = tile_indice_path / prod.identifier

                            for fid in fid_list:
                                df_dict = OrderedDict()
                                df_dict["fid"] = fid
                                # feat_properties = features[fid]["properties"]
                                # if feat_properties:
                                    # df_dict.update(feat_properties)
                                df_dict.update(TimeSeries._get_raster_stats_in_geom(features[fid], prod_path))

                                # df_properties = features[fid]["properties"]
                                df_dict["date"] = prod.date
                                df_dict["tile"] = tile
                                df_dict["filename"] = prod.identifier

                                rows_list.append(df_dict)

                self._df_dicts[indice] = TimeSeries._list_to_df(rows_list)

    @staticmethod
    def _list_to_df(rows_list: List[Dict]) -> pd.DataFrame:
        """Converts a list of dictionaries as a pandas dataframe.

        :param rows_list: list of dictionaries.
        """
        df = pd.DataFrame.from_dict(rows_list)
        df = df.sort_values(by=["date"])
        df.set_index("date", inplace=True, drop=False)
        del df['date']
        return df

    def to_csv(self, out_path: Union[str, pathlib.PosixPath] = None) -> None:
        """Exports the time series to CSV format.

        :param out_path: output folder. Default is DATA/TIME_SERIES.
        """
        if out_path is None:
            out_path = Config().get("time_series_path")
        for df_name, df in self._df_dicts.items():
            csv_path = Path(out_path) / "{0}_{1}_{2}_{3}.csv".format(
                self._vectors_file.stem,
                df_name,
                self._date_min,
                self._date_max)
            df.to_csv(str(csv_path))
            logger.info("exported to csv: {}".format(csv_path))
